#Copyright 2021
#Author Dallin Skouson
#see the license for details
#
#Tests the verilog composers functions and output

from collections import OrderedDict
import unittest
from unittest.case import expectedFailure
import spydrnet as sdn
from spydrnet.composers.verilog.composer import Composer
from collections import OrderedDict

class TestVerilogComposerUnit(unittest.TestCase):

    class TestFile:
        '''represents a file (has a write function for the composer)
        can be used as a drop in replacement for the composer file.write function
        saves all written stuff to a string'''
        def __init__(self):
            self.written = ""

        def write(self, text):
            self.written += text

        def clear(self):
            self.written = ""

        def compare(self, text, should_match = True):
            self.written = self.written.lstrip()
            if (text == self.written) == should_match:
                return True
            else:
                print("The composer wrote:")
                print('"' + self.written + '"')
                print("This was compared to:")
                print('"' + text + '"')
                if not should_match:
                    print("and these are not supposed to match")
                else:
                    print("and these should have matched")
                print("\n")
                return False


    def initialize_tests(self):
        composer = Composer()
        composer.file = self.TestFile()
        return composer

    def initialize_netlist(self):
        netlist = sdn.Netlist()
        netlist.name = "test_netlist"
        return netlist

    def initialize_library(self):
        netlist = self.initialize_netlist()
        library = netlist.create_library()
        library.name = "test_library"
        return library

    def initialize_definition(self):
        library = self.initialize_library()
        definition = library.create_definition()
        definition.name = "test_definition"
        return definition

    def initialize_instance_parameters(self, instance):
        instance["VERILOG.Parameters"] = OrderedDict()
        instance["VERILOG.Parameters"]["key"] = "value"
        instance["VERILOG.Parameters"]["key2"] = "value2"

        expected1 = ".key(value)"
        expected2 = ".key2(value2)"

        return expected1, expected2

    def initialize_instance_port_connections(self, instance, definition):
        ref_def = definition.library.create_definition()
        instance.reference = ref_def
        ref_def.name = "reference_definition"

        single_bit_port = ref_def.create_port()
        single_bit_port.create_pin()
        single_bit_port.is_downto = True
        single_bit_port.name = "single_bit_port"

        single_bit_cable = definition.create_cable()
        single_bit_cable.name = "single_bit_cable"
        single_bit_cable.is_downto = True
        single_bit_cable.create_wire()


        multi_bit_port = ref_def.create_port()
        multi_bit_port.is_downto = True
        multi_bit_port.create_pins(4)
        multi_bit_port.name = "multi_bit_port"

        multi_bit_port_offset = ref_def.create_port()
        multi_bit_port_offset.lower_index = 4
        multi_bit_port_offset.is_downto = True
        multi_bit_port_offset.create_pins(4)
        multi_bit_port_offset.name = "multi_bit_port_offset"

        partial_port = ref_def.create_port()
        partial_port.create_pins(2)
        partial_port.is_downto = True
        partial_port.name = "partial_port"

        multi_bit_cable = definition.create_cable()
        multi_bit_cable.create_wires(4)
        multi_bit_cable.name = "multi_bit_cable"
        multi_bit_cable.is_downto = True


        concatenated_port = ref_def.create_port()
        concatenated_port.create_pins(4)
        concatenated_port.name = "concatenated_port"

        ccs = []
        for i in range(4):
            cable = definition.create_cable()
            cable.create_wire()
            cable.is_downto = True
            cable.name = "cc_" + str(i)
            ccs.append(cable)


        single_bit_cable.wires[0].connect_pin(instance.pins[single_bit_port.pins[0]])

        for i in range(4):
            multi_bit_cable.wires[i].connect_pin(instance.pins[multi_bit_port.pins[i]])
            multi_bit_cable.wires[i].connect_pin(instance.pins[multi_bit_port_offset.pins[i]])
            ccs[i].wires[0].connect_pin(instance.pins[concatenated_port.pins[i]])

        for i in range(2):
            multi_bit_cable.wires[i].connect_pin(instance.pins[partial_port.pins[i]])

        single_bit_expected = "." + single_bit_port.name + "(" + single_bit_cable.name + ")"

        multi_bit_expected = "." + multi_bit_port.name + "(" + multi_bit_cable.name + ")"

        offset_expected = "." + multi_bit_port_offset.name + "(" + multi_bit_cable.name + ")"

        partial_expected = "." + partial_port.name + "(" + multi_bit_cable.name + "[1:0])"

        concatenated_expected = "." + concatenated_port.name + "({" + ccs[0].name + ', ' + ccs[1].name + ', ' + ccs[2].name + ', ' + ccs[3].name + "})"

        return single_bit_port, single_bit_expected, \
            multi_bit_port, multi_bit_expected, \
            multi_bit_port_offset, offset_expected, \
            partial_port, partial_expected,\
            concatenated_port, concatenated_expected\

    def test_write_header(self):
        composer = self.initialize_tests()
        netlist = sdn.Netlist()
        netlist.name = "Netlist_name"
        composer._write_header(netlist)
        assert composer.file.compare("//Generated from netlist by SpyDrNet\n//netlist name: Netlist_name\n")

    def test_write_brackets_single_bit(self):
        #def _write_brackets(self, bundle, low_index, high_index):
        composer = self.initialize_tests()

        port = sdn.Port()
        cable = sdn.Cable()

        cable_name = "my_cable"
        port_name = "my_port"

        port.name = port_name
        cable.name = cable_name

        port.create_pin()
        cable.create_wire()

        composer._write_brackets(port, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 0, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, None, 0)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 0, 0)
        assert composer.file.compare("")
        composer.file.clear()

        composer._write_brackets(cable, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 0, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, None, 0)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 0, 0)
        assert composer.file.compare("")
        composer.file.clear()
        #none of these should write because they are all single bit.

    def test_write_brackets_single_bit_offset(self):
        #def _write_brackets(self, bundle, low_index, high_index):
        composer = self.initialize_tests()

        port = sdn.Port()
        cable = sdn.Cable()

        cable_name = "my_cable"
        port_name = "my_port"

        port.name = port_name
        cable.name = cable_name

        port.create_pin()
        cable.create_wire()

        port.lower_index = 4
        cable.lower_index = 4

        composer._write_brackets(port, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 4, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, None, 4)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 4, 4)
        assert composer.file.compare("")
        composer.file.clear()

        composer._write_brackets(cable, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 4, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, None, 4)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 4, 4)
        assert composer.file.compare("")
        composer.file.clear()
        #none of these should write because they are all single bit.

    def test_write_brackets_multi_bit(self):
        composer = self.initialize_tests()

        port = sdn.Port()
        cable = sdn.Cable()

        cable_name = "my_cable"
        port_name = "my_port"

        port.name = port_name
        cable.name = cable_name

        port.create_pins(4) #input [3:0] my_input;
        port.is_downto = True
        cable.create_wires(4) #wire [3:0] my_wire;
        cable.is_downto = True

        composer._write_brackets(port, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 1, None)
        assert composer.file.compare("[1]")
        composer.file.clear()
        composer._write_brackets(port, None, 2)
        assert composer.file.compare("[2]")
        composer.file.clear()
        composer._write_brackets(port, 2, 2)
        assert composer.file.compare("[2]")
        composer.file.clear()
        composer._write_brackets(port, 0, 3)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 1, 2)
        assert composer.file.compare("[2:1]")
        composer.file.clear()

        composer._write_brackets(cable, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 1, None)
        assert composer.file.compare("[1]")
        composer.file.clear()
        composer._write_brackets(cable, None, 2)
        assert composer.file.compare("[2]")
        composer.file.clear()
        composer._write_brackets(cable, 2, 2)
        assert composer.file.compare("[2]")
        composer.file.clear()
        composer._write_brackets(cable, 0, 3)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 1, 2)
        assert composer.file.compare("[2:1]")
        composer.file.clear()


    def test_write_brackets_multi_bit_offset(self):
        composer = self.initialize_tests()

        port = sdn.Port()
        cable = sdn.Cable()

        cable_name = "my_cable"
        port_name = "my_port"

        port.name = port_name
        cable.name = cable_name

        port.create_pins(4) #input [3:0] my_input;
        port.is_downto = True
        port.lower_index = 4
        cable.create_wires(4) #wire [3:0] my_wire;
        cable.is_downto = True
        cable.lower_index = 4

        composer._write_brackets(port, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 5, None)
        assert composer.file.compare("[5]")
        composer.file.clear()
        composer._write_brackets(port, None, 6)
        assert composer.file.compare("[6]")
        composer.file.clear()
        composer._write_brackets(port, 6, 6)
        assert composer.file.compare("[6]")
        composer.file.clear()
        composer._write_brackets(port, 4, 7)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(port, 5, 6)
        assert composer.file.compare("[6:5]")
        composer.file.clear()

        composer._write_brackets(cable, None, None)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 5, None)
        assert composer.file.compare("[5]")
        composer.file.clear()
        composer._write_brackets(cable, None, 6)
        assert composer.file.compare("[6]")
        composer.file.clear()
        composer._write_brackets(cable, 6, 6)
        assert composer.file.compare("[6]")
        composer.file.clear()
        composer._write_brackets(cable, 4, 7)
        assert composer.file.compare("")
        composer.file.clear()
        composer._write_brackets(cable, 5, 6)
        assert composer.file.compare("[6:5]")
        composer.file.clear()

    def test_write_brackets_fail(self):
        pass #we should add some tests to test out of bounds on the brackets.

    def test_write_brackets_defining(self):

        composer = self.initialize_tests()

        def initialize_bundle(bundle, offset, width):
            if isinstance(bundle, sdn.Port):
                bundle.create_pins(width)
            else: #it's a cable
                bundle.create_wires(width)
            bundle.is_downto = True
            bundle.lower_index = offset
            return bundle

        b1 = initialize_bundle(sdn.Port(), 0, 1)
        b2 = initialize_bundle(sdn.Cable(), 4, 1)
        b3 = initialize_bundle(sdn.Port(), 0, 4)
        b4 = initialize_bundle(sdn.Cable(), 4, 4)

        composer._write_brackets_defining(b1)
        assert composer.file.compare("")
        composer.file.clear()

        composer._write_brackets_defining(b2)
        assert composer.file.compare("[4:4]")
        composer.file.clear()

        composer._write_brackets_defining(b3)
        assert composer.file.compare("[3:0]")
        composer.file.clear()

        composer._write_brackets_defining(b4)
        assert composer.file.compare("[7:4]")
        composer.file.clear()

    def test_write_name(self):
        composer = self.initialize_tests()
        o = sdn.Cable() #Type of this shouldn't really matter
        valid_names = ["basic_name", "\\escaped ", "\\fads#@%!$!@#%$[0:4320] "]
        for n in valid_names:
            o.name = n
            composer._write_name(o)
            assert composer.file.compare(n)
            composer.file.clear()

    @unittest.expectedFailure
    def test_write_none_name(self):
        composer = self.initialize_tests()
        o = sdn.Cable()
        composer._write_name(o)

    @unittest.expectedFailure
    def test_write_invalid_name(self):
        composer = self.initialize_tests()
        o = sdn.Cable()
        o.name = "\\escaped_no_space"
        composer._write_name(o)

    def test_write_instance_port(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()
        instance = definition.create_child()
        instance.name = "ports_test"

        single_bit_port, single_bit_expected, \
            multi_bit_port, multi_bit_expected, \
            multi_bit_port_offset, offset_expected, \
            partial_port, partial_expected, \
            concatenated_port, concatenated_expected\
            = self.initialize_instance_port_connections(instance, definition)

        composer._write_instance_port(instance, single_bit_port)
        assert composer.file.compare(single_bit_expected)
        composer.file.clear()

        composer._write_instance_port(instance, multi_bit_port)
        assert composer.file.compare(multi_bit_expected)
        composer.file.clear()

        composer._write_instance_port(instance, multi_bit_port_offset)
        assert composer.file.compare(offset_expected)
        composer.file.clear()

        composer._write_instance_port(instance, partial_port)
        assert composer.file.compare(partial_expected)
        composer.file.clear()

        composer._write_instance_port(instance, concatenated_port)
        assert composer.file.compare(concatenated_expected)
        composer.file.clear()

        composer._write_instance_ports(instance)
        expected = "(\n"
        first = True
        expected_strs = [single_bit_expected, multi_bit_expected, offset_expected, partial_expected, concatenated_expected]
        for i in expected_strs:
            if not first:
                expected += ",\n"
            expected +=  "        "
            expected += i
            first = False
        expected += "\n    );"
        assert composer.file.compare(expected)

    def test_write_instance_parameters(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()
        instance = definition.create_child()
        instance.name = "ports_test"
        ref_def = definition.library.create_definition()
        instance.reference = ref_def

        expected1, expected2 =self.initialize_instance_parameters(instance)
        #instance["VERILOG.Parameters"]["no_value"] = None #always has value?

        composer._write_instance_parameter("key", "value")
        assert composer.file.compare(expected1)
        composer.file.clear()

        composer._write_instance_parameter("key2", "value2")
        assert composer.file.compare(expected2)
        composer.file.clear()


        # composer._write_instance_parameter("no_value", None)
        # expected3 = ".key()"
        # assert composer.file.compare(expected2)
        # composer.file.clear()

        composer._write_instance_parameters(instance)
        expected = "#(\n        " + expected1 + ",\n        " + expected2 + "\n    )\n"
        assert composer.file.compare(expected)


    def test_write_full_instance(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()
        instance = definition.create_child()
        instance.name = "instance_test"

        expected1, expected2 = self.initialize_instance_parameters(instance)
        parameters_expected = "#(\n        " + expected1 + ",\n        " + expected2 + "\n    )\n"
        single_bit_port, single_bit_expected, \
            multi_bit_port, multi_bit_expected, \
            multi_bit_port_offset, offset_expected, \
            partial_port, partial_expected, \
            concatenated_port, concatenated_expected\
            = self.initialize_instance_port_connections(instance, definition)

        port_expected = "\n    (\n"
        first = True
        expected_strs = [single_bit_expected, multi_bit_expected, offset_expected, partial_expected, concatenated_expected]
        for i in expected_strs:
            if not first:
                port_expected += ",\n"
            port_expected +=  "        "
            port_expected += i
            first = False
        port_expected += "\n    );"

        composer._write_module_body_instance(instance)

        expected = instance.reference.name + " " + parameters_expected + "    " + instance.name + port_expected + "\n"
        assert composer.file.compare(expected)

    def test_write_module_header(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()
        definition["VERILOG.Parameters"] = OrderedDict()

        definition["VERILOG.Parameters"]["key"] = "value"
        definition["VERILOG.Parameters"]["no_default"] = None

        port = definition.create_port()
        port.name = "my_port"
        port.create_pin()

        port = definition.create_port()
        port.name = "my_port2"
        port.create_pin()

        port.direction = sdn.Port.Direction.IN

        composer._write_module_header(definition)
        expected = "module " + definition.name + "\n#(\n    parameter key = value,\n    parameter no_default\n)(\n    my_port,\n    my_port2\n);\n\n"
        assert composer.file.compare(expected)

    def test_write_module_ports_header_and_body_alias(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()

        port_alias = definition.create_port()
        port_alias.name = "aliased"
        port_alias.create_pins(2)
        port_alias.direction = sdn.Port.Direction.IN

        c1 = definition.create_cable("c1")
        c2 = definition.create_cable("c2")
        c1.create_wire()
        c2.create_wire()

        c1.wires[0].connect_pin(port_alias.pins[0])
        c2.wires[0].connect_pin(port_alias.pins[1])

        composer._write_module_header_port(port_alias)
        assert composer.file.compare("." + port_alias.name + "({"+ c1.name + ", " + c2.name +"})")
        composer.file.clear()
        composer._write_module_body_port(port_alias)
        assert composer.file.compare("input " + c1.name + ";\n    " + "input " + c2.name + ";\n")
        composer.file.clear()


    def test_write_module_ports_header_and_body_multi(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()

        port_multi = definition.create_port("multi_bit")
        port_multi.name = "multi_bit"
        port_multi.create_pins(4)
        port_multi.direction = sdn.Port.Direction.OUT

        cable_multi = definition.create_cable("multi_bit")
        cable_multi.create_wires(4)

        for i in range(4):
            cable_multi.wires[i].connect_pin(port_multi.pins[i])

        composer._write_module_header_port(port_multi)
        assert composer.file.compare(port_multi.name)
        composer.file.clear()
        composer._write_module_body_port(port_multi)
        assert composer.file.compare("output [3:0]" + port_multi.name + ";\n")
        composer.file.clear()


    def test_write_module_ports_header_and_body_disconnect(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()

        port_disconnect = definition.create_port("disconnected")
        port_disconnect.direction = sdn.Port.Direction.INOUT
        port_disconnect.create_pin()

        composer._write_module_header_port(port_disconnect)
        assert composer.file.compare(port_disconnect.name)
        composer.file.clear()
        composer._write_module_body_port(port_disconnect)
        assert composer.file.compare("inout " + port_disconnect.name + ';\n')
        composer.file.clear()

    def test_write_module_body_cables(self):
        composer = self.initialize_tests()
        definition = self.initialize_definition()

        cable = definition.create_cable(name = "test_cable", is_downto = True)
        cable.create_wires(4)

        composer._write_module_body_cable(cable)
        assert composer.file.compare("wire [3:0]" + cable.name + ";\n")


    def test_assignment_single_bit(self):
        pass

    def test_assignment_multi_bit(self):
        pass




import unittest
import spydrnet as sdn


class TestGetInstances(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        instance = definition.create_child()
        instance.name = "MY_INST"
        top_instance = sdn.Instance()
        top_instance.name = "TOP_INST"
        top_instance.reference = definition
        netlist.top_instance = top_instance
        cls.netlist = netlist
        cls.library = library
        cls.definition = definition
        cls.instance = instance
        cls.top_instance = top_instance

        leaf_definition = library.create_definition()
        leaf_cable = leaf_definition.create_cable()
        leaf_wire = leaf_cable.create_wire()
        instance.reference = leaf_definition
        cls.leaf_definition = leaf_definition
        cls.leaf_cable = leaf_cable
        cls.leaf_wire = leaf_wire

        cable = definition.create_cable()
        wire = cable.create_wire()
        port = definition.create_port()
        pin = port.create_pin()
        outer_pin = top_instance.pins[pin]
        cls.cable = cable
        cls.wire = wire
        cls.port = port
        cls.pin = pin
        cls.outer_pin = outer_pin

    def test_parameter_checking(self):
        definition = sdn.Definition()
        instance = definition.create_child()
        instance.name = "MY_INST"
        self.assertRaises(TypeError, sdn.get_instances, definition, "MY_INST", patterns="MY_INST")
        self.assertRaises(TypeError, sdn.get_instances, definition, "MY_INST", unsupported_keyword=None)
        self.assertRaises(TypeError, sdn.get_instances, definition, "MY_INST", selection=sdn.BOTH)
        self.assertRaises(TypeError, sdn.get_instances, None, "MY_INST")
        self.assertRaises(TypeError, sdn.get_instances, [None, definition], "MY_INST")

    def test_collection(self):
        instances = list(sdn.get_instances([self.definition, self.definition]))
        self.assertEqual(len(instances), 1)

    def test_get_instances_in_library(self):
        instance1 = next(self.library.get_instances("MY_INST"))
        self.assertEqual(self.instance, instance1)

    def test_get_instances_in_netlist(self):
        instance1 = next(self.netlist.get_instances("MY_INST"))
        self.assertEqual(self.instance, instance1)

    def test_get_instances_from_netlist(self):
        instance1 = next(sdn.get_instances(self.netlist))
        self.assertEqual(self.instance, instance1)

    def test_get_instances_from_library(self):
        instance1 = next(sdn.get_instances(self.library))
        self.assertEqual(self.instance, instance1)

    def test_get_instances_from_definition_recursive(self):
        instance1 = list(sdn.get_instances(self.definition, recursive=True))
        self.assertEqual(self.instance, instance1[0])

    def test_from_definition_outside_recursive(self):
        instance1 = list(sdn.get_instances(self.definition, selection="OUTSIDE", recursive=True))
        self.assertEqual(self.top_instance, instance1[0])

    def test_from_leaf_definition_outside_recursive(self):
        instance1 = list(sdn.get_instances(self.leaf_definition, selection="OUTSIDE", recursive=True))
        self.assertEqual(len(instance1), 2)
        self.assertTrue(self.top_instance in instance1 and self.instance in instance1)

    def test_from_instance_recursive(self):
        instance1 = next(sdn.get_instances(self.top_instance, recursive=True))
        self.assertEqual(self.instance, instance1)

    def test_from_instance_outside_resursive(self):
        instance1 = list(sdn.get_instances(self.instance, selection="OUTSIDE", recursive=True))
        self.assertEqual(self.top_instance, instance1[0])

    def test_from_wire(self):
        instance1 = next(sdn.get_instances(self.wire))
        self.assertIs(self.top_instance, instance1)

    def test_from_outer_pin(self):
        instance1 = next(sdn.get_instances(self.outer_pin, "TOP_INST"))
        self.assertIs(self.top_instance, instance1)

    def test_from_inner_pin(self):
        instance1 = next(sdn.get_instances(self.pin, "TOP_INST"))
        self.assertIs(self.top_instance, instance1)

    def test_from_href_of_instance(self):
        top_href = next(sdn.get_hinstances(self.netlist.top_instance))
        instance1 = next(sdn.get_instances(top_href))
        self.assertIs(self.top_instance, instance1)

    def test_from_href_of_port(self):
        port_href = next(sdn.get_hports(self.port))
        instance1 = next(sdn.get_instances(port_href))
        self.assertIs(self.top_instance, instance1)



import unittest
import spydrnet as sdn
from spydrnet.composers.edif.composer import ComposeEdif


class TestTopologicalSort(unittest.TestCase):

    def test_top_sort(self):
        ce = ComposeEdif()
        def library_dependence(library):
            depend_set = set()
            for definition in library.definitions:
                # print("library: " + library.name)
                for child in definition.children:
                    # print("contains: " + child.name)
                    if child.reference.library != library:
                        # print("which instances definition " + child.reference.name)
                        # print("which depends on " + child.reference.library.name)
                        depend_set.add(child.reference.library)
            # print("DEPENDENCY LIST RETURNED")
            # for lib_dep in depend_set:
            #     print(lib_dep.name)
            return depend_set

        netlist = sdn.Netlist()

        lib_a = netlist.create_library()
        lib_a.name = "lib_a"
        definition_a = lib_a.create_definition()
        definition_a.name = "def_a"

        lib_b = netlist.create_library()
        lib_b.name = "lib_b"
        definition_b = lib_b.create_definition()
        definition_b.name = "def_b"
        
        instance_b = definition_b.create_child()
        instance_b.name = "relies_on_lib_a"
        instance_b.reference = definition_a

        lib_c = netlist.create_library()
        lib_c.name = "lib_c"
        definition_c = lib_c.create_definition()
        definition_c.name = "def_c"

        instance_c1 = definition_c.create_child()
        instance_c1.name = "relies_on_a"
        instance_c1.reference = definition_a
        instance_c2 = definition_c.create_child()
        instance_c2.name = "relies_on_b"
        instance_c2.reference = definition_b

        libraries = ce._topological_sort(netlist.libraries, library_dependence)

        print("should be a, b, c")
        for lib in libraries:
            print(lib.name)

        print("dependencies")
        for lib in libraries:
            print(lib.name, end = "")
            print(" relies on")
            for l in library_dependence(lib):
                print("\t" + l.name)

        for l in libraries:
            assert l in netlist.libraries

        for l in netlist.libraries:
            assert l in libraries

        found_libraries = set()
        for l in libraries:
            found_libraries.add(l)
            for l_dep in library_dependence(l):
                assert l_dep in found_libraries


import unittest
import spydrnet as sdn


class TestGetHInstances(unittest.TestCase):
    netlist = None
    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.load_example_netlist_by_name('b13')

    def test_absolute_search(self):
        href = next(self.netlist.get_hinstances('tx_end_reg'), None)
        self.assertIsNotNone(href)

    def test_regex_search(self):
        hrefs = list(sdn.get_hinstances(self.netlist, '.*FSM_onehot.*', is_re=True))
        self.assertTrue(len(hrefs) == 24)

    def test_parameters(self):
        self.assertRaises(TypeError, sdn.get_hinstances, self.netlist, r'.*FSM_onehot.*', patterns=r".*FSM_onehot.*")
        self.assertRaises(TypeError, sdn.get_hinstances, self.netlist, parameter_does_not_exit=True)
        self.assertRaises(TypeError, sdn.get_hinstances, object())

    def test_get_hinstances_of_pin(self):
        library = self.netlist.libraries[0]
        definition = library.definitions[0]
        port = definition.ports[0]
        pin = port.pins[0]
        hrefs = list(sdn.get_hinstances(pin))
        self.assertTrue(len(hrefs) == 7)

    def test_get_hinstances_of_wire(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hinstances(wire))
        self.assertTrue(len(hrefs) == 1)

    def test_get_hinstance_of_invalid_reference(self):
        from spydrnet.util.hierarchical_reference import HRef
        invalid_href = HRef.from_parent_and_item(None, None)
        hrefs = list(sdn.get_hinstances(invalid_href))
        self.assertTrue(len(hrefs) == 0)

    def test_get_hinstances_from_hrefs_of_cable_and_wire(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hinstances(wire))
        href_top = hrefs[0]
        from spydrnet.util.hierarchical_reference import HRef
        cable_href = HRef.from_parent_and_item(href_top, cable)
        href_result = next(sdn.get_hinstances(cable_href), None)
        self.assertTrue(href_result is href_top)
        wire_href = HRef.from_parent_and_item(cable_href, wire)
        href_result = next(sdn.get_hinstances(wire_href), None)
        self.assertTrue(href_result is href_top)

    def test_get_hinstances_from_hrefs_of_port_and_pin(self):
        library = self.netlist.libraries[0]
        definition = library.definitions[0]
        port = definition.ports[0]
        pin = port.pins[0]
        hrefs = list(sdn.get_hinstances(pin))
        href = hrefs[0]
        from spydrnet.util.hierarchical_reference import HRef
        port_href = HRef.from_parent_and_item(href, port)
        href_result = next(sdn.get_hinstances(port_href), None)
        self.assertTrue(href_result is href)
        pin_href = HRef.from_parent_and_item(port_href, pin)
        href_result = next(sdn.get_hinstances(pin_href), None)
        self.assertTrue(href_result is href)

    def test_from_href_of_instance(self):
        href = next(sdn.get_hinstances(self.netlist.top_instance))
        hrefs = list(sdn.get_hinstances(href))
        self.assertTrue(len(hrefs) == 102)

    def test_from_library(self):
        hrefs = list(sdn.get_hinstances(self.netlist.libraries[0]))
        self.assertTrue(len(hrefs) == 102)

    def test_from_wire_and_cable(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hinstances(self.netlist.top_instance))
        href_top = hrefs[0]
        href_result = next(sdn.get_hinstances(wire))
        self.assertTrue(href_top is href_result)
        href_result = next(sdn.get_hinstances(cable))
        self.assertTrue(href_top is href_result)

    def test_from_outerpin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        instance = definition.children[0]
        outerpin = next(iter(instance.pins))
        hrefs = list(sdn.get_hinstances(outerpin))
        self.assertTrue(len(hrefs) == 1)

    def test_of_bad_instance(self):
        hrefs = list(sdn.get_hinstances(sdn.Instance()))
        self.assertTrue(len(hrefs) == 0)

    def test_through_hierarchy(self):
        netlist = sdn.Netlist()
        library = netlist.create_library()
        definition = library.create_definition()
        instance = sdn.Instance()
        instance.reference = definition
        netlist.top_instance = instance

        middle_inst = definition.create_child()
        middle_def = library.create_definition()
        middle_inst.reference = middle_def

        leaf_inst = middle_def.create_child()
        leaf_def = library.create_definition()
        leaf_inst.reference = leaf_def

        hrefs = list(sdn.get_hinstances(leaf_inst))
        self.assertTrue(len(hrefs) == 1)

    @unittest.skip("Test takes too long at this time.")
    def test_recursive_memory_use(self):
        netlist = sdn.load_example_netlist_by_name('leon3mp_hierarchical')
        hrefs = list(sdn.get_hinstances(netlist))
        self.assertTrue(len(hrefs) > 0)



import unittest
import spydrnet as sdn

class TestGlobalServiceWithoutPlugins(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        sdn.namespace_manager.deregister_all_listeners()
        cls.netlist = sdn.load_example_netlist_by_name("b13")

    @classmethod
    def tearDownClass(cls) -> None:
        sdn.namespace_manager.register_all_listeners()

    def test_double_register_lookup(self):
        from spydrnet.global_state.global_service import register_lookup
        register_lookup("TEST", None)
        self.assertRaises(ValueError, register_lookup, "TEST", None)

    def test_lookup_library(self):
        library = self.netlist.libraries[0]
        library_name = library.name
        lib1 = next(self.netlist.get_libraries(library_name))
        self.assertEqual(library, lib1)
        lib2 = next(self.netlist.get_libraries(library_name, key='EDIF.identifier'))
        self.assertEqual(library, lib2)
        lib3 = next(sdn.get_libraries(self.netlist, library_name))
        self.assertEqual(library, lib3)
        lib4 = next(sdn.get_libraries(self.netlist, library_name, key='EDIF.identifier'))
        self.assertEqual(library, lib4)

    def test_lookup_definition(self):
        library = self.netlist.libraries[0]
        definition = library.definitions[0]
        definition_name = definition.name
        def1 = next(self.netlist.get_definitions(definition_name))
        self.assertEqual(definition, def1)
        def2 = next(self.netlist.get_definitions(definition_name, key='EDIF.identifier'))
        self.assertEqual(definition, def2)
        def3 = next(sdn.get_definitions(self.netlist, definition_name))
        self.assertEqual(definition, def3)
        def4 = next(sdn.get_definitions(self.netlist, definition_name, key='EDIF.identifier'))
        self.assertEqual(definition, def4)
        def5 = next(library.get_definitions(definition_name))
        self.assertEqual(definition, def5)
        def6 = next(library.get_definitions(definition_name, key='EDIF.identifier'))
        self.assertEqual(definition, def6)
        def7 = next(sdn.get_definitions(library, definition_name))
        self.assertEqual(definition, def7)
        def8 = next(sdn.get_definitions(library, definition_name, key='EDIF.identifier'))
        self.assertEqual(definition, def8)

    def test_lookup_port_cables_instances(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        port = definition.ports[0]
        port_name = port.name
        cable = definition.cables[0]
        cable_name = cable.name
        instance = definition.children[0]
        instance_name = instance.name
        port1 = next(definition.get_ports(port_name))
        self.assertEqual(port, port1)
        cable1 = next(definition.get_cables(cable_name))
        self.assertEqual(cable, cable1)
        instance1 = next(definition.get_instances(instance_name))
        self.assertEqual(instance, instance1)

    def test_lookup_something_that_does_not_exist(self):
        item = next(self.netlist.get_libraries("DOES_NOT_EXIST"), None)
        self.assertIsNone(item)



import spydrnet.parsers.verilog.verilog_tokens as vt

class TokenFactory():
    def __init__(self):    
        self.buffer = ''
        #only one of these should ever be true at any time.
        self.single_line_comment = False
        self.multi_line_comment = False
        self.string = False
        self.escaped_identifier = False
        self.processor_directive = False
        self.last_character = "" #should not be relied on to always be the last character when tokens are ending
    
    def clear(self, character = ""):
        self.buffer = character

    def any_flag_set(self):
        return self.single_line_comment or self.multi_line_comment or self.string or self.escaped_identifier or self.processor_directive

    def set_flags(self):
        if len(self.buffer) <= 2 and not self.any_flag_set():
            
            if self.buffer == vt.OPEN_LINE_COMMENT:
                self.single_line_comment = True
        
            elif self.buffer == vt.OPEN_BLOCK_COMMENT:
                self.multi_line_comment = True
        
            elif self.buffer == '"':
                self.string = True
        
            elif self.buffer == "\\":
                self.escaped_identifier = True
            
            elif self.buffer == "`":
                self.processor_directive = True

    def add_character(self, character):
        #tries to add the character to the token.
        #if the token is complete it will return and start the next token
        #if the token is incomplete it will add it and return none
        token_out = None
        #first thing, see if we need to return a token
        if self.buffer in vt.SINGLE_CHARACTER_TOKENS:
            token_out = self.buffer
            self.clear()

        elif self.single_line_comment and character == vt.NEW_LINE:
            token_out = self.buffer
            self.clear()
            self.single_line_comment = False

        elif self.multi_line_comment and self.last_character == "*" and character == "/":
            token_out = self.buffer + character
            character = ""
            self.clear()
            self.multi_line_comment = False

        elif self.escaped_identifier and character in vt.WHITESPACE:
            token_out = self.buffer + " "
            self.clear()
            self.escaped_identifier = False

        elif self.string and character == '"':
            token_out = self.buffer + character
            character = ""
            self.clear()
            self.string = False

        elif self.processor_directive and character == vt.NEW_LINE:
            token_out = self.buffer
            self.clear()
            self.processor_directive = False

        elif character == vt.STAR and self.last_character == "/":
            pass #this will be a multi line comment but * is a breaker...

        elif character in vt.BREAKER_TOKENS and not self.any_flag_set() and len(self.buffer) != 0:
            token_out = self.buffer
            self.clear()

        elif character == vt.DOT and not self.any_flag_set() and len(self.buffer) != 0 and\
                not vt.is_numeric(self.buffer):
            token_out = self.buffer
            self.clear()

        
        if character not in vt.WHITESPACE:
            self.buffer = self.buffer + character
        elif self.single_line_comment or self.multi_line_comment or self.string or self.processor_directive:
            self.buffer = self.buffer + character
        self.last_character = character
        
        last_multi_line_comment = self.multi_line_comment
        self.set_flags()
        if self.multi_line_comment and not last_multi_line_comment:
            self.last_character = "" #these lines help to ensure that /*/ isn't precieved as a complete comment.


        return token_out

    def flush(self):
        token_out = self.buffer
        self.clear()
        if token_out == "":
            return None
        return token_out


import unittest
import spydrnet as sdn


class TestGetHPorts(unittest.TestCase):
    netlist = None
    @classmethod
    def setUpClass(cls) -> None:
        cls.netlist = sdn.load_example_netlist_by_name('b13')

    def test_wild_card_search_on_netlist(self):
        hrefs = list(sdn.get_hports(self.netlist))
        self.assertTrue(len(hrefs) == 12 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_wild_card_search_on_netlist_recursive(self):
        hrefs = list(sdn.get_hports(self.netlist, recursive=True))
        self.assertTrue(len(hrefs) == 500 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_absolute_search_excluded(self):
        href = next(self.netlist.get_hports('this_is_not_contained'), None)
        self.assertIsNone(href)

    def test_absolute_search(self):
        href = next(self.netlist.get_hports('clock'), None)
        self.assertIsNotNone(href)
        self.assertIsInstance(href.item, sdn.Port)

    def test_regex_search(self):
        hrefs = list(sdn.get_hports(self.netlist, '.*dAtA.*', is_re=True, is_case=False))
        self.assertTrue(len(hrefs) == 2 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_parameters(self):
        self.assertRaises(TypeError, sdn.get_hports, self.netlist, r'.*FSM_onehot.*', patterns=r".*FSM_onehot.*")
        self.assertRaises(TypeError, sdn.get_hports, self.netlist, parameter_does_not_exit=True)
        self.assertRaises(TypeError, sdn.get_hports, object())

    def test_get_hports_of_pin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        port = definition.ports[0]
        pin = port.pins[0]
        hrefs = list(sdn.get_hports(pin))
        self.assertTrue(len(hrefs) == 1)
        self.assertIsInstance(hrefs[0].item, sdn.Port)

    def test_get_hports_of_wire(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hports(wire))
        self.assertTrue(len(hrefs) == 2)
        self.assertIsInstance(hrefs[0].item, sdn.Port)

    def test_get_hwire_of_invalid_reference(self):
        from spydrnet.util.hierarchical_reference import HRef
        invalid_href = HRef.from_parent_and_item(None, None)
        hrefs = list(sdn.get_hports(invalid_href))
        self.assertTrue(len(hrefs) == 0)

    def test_get_hpin_from_href_pin(self):
        pin = self.netlist.libraries[1].definitions[0].ports[0].pins[0]
        href = next(pin.get_hports())
        href_from_href = next(sdn.get_hports(href))
        self.assertIs(href, href_from_href)

    def test_get_hports_from_hrefs_of_cable_and_wire(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        cable_href = next(cable.get_hcables())
        wire_href = next(wire.get_hwires())
        href_cable_pins = set(sdn.get_hports(cable_href))
        href_wire_pins = set(sdn.get_hports(wire_href))
        self.assertTrue(len(href_cable_pins) > 0 and len(href_wire_pins) > 0)
        self.assertTrue(all(isinstance(x.item, sdn.Port) for x in href_cable_pins))
        self.assertTrue(all(isinstance(x.item, sdn.Port) for x in href_wire_pins))
        self.assertTrue(all(x in href_cable_pins for x in href_wire_pins))

    def test_get_hports_from_hrefs_of_port_and_pin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        port = definition.ports[0]
        pin = port.pins[0]
        hrefs = list(sdn.get_hinstances(pin))
        href = hrefs[0]
        from spydrnet.util.hierarchical_reference import HRef
        port_href = HRef.from_parent_and_item(href, port)
        pin_href = HRef.from_parent_and_item(port_href, pin)
        href_result = next(sdn.get_hports(port_href), None)
        self.assertTrue(href_result is port_href)
        href_result = next(sdn.get_hports(pin_href), None)
        self.assertTrue(href_result is port_href)

    def test_from_href_of_instance(self):
        href = next(sdn.get_hinstances(self.netlist.top_instance))
        hrefs = list(sdn.get_hports(href))
        self.assertTrue(len(hrefs) == 12 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_from_instance(self):
        hrefs = list(sdn.get_hports(self.netlist.top_instance))
        self.assertTrue(len(hrefs) == 12 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_from_library(self):
        hrefs = list(sdn.get_hports(self.netlist.libraries[0]))
        self.assertTrue(len(hrefs) == 488 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_from_wire_and_cable(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        cable = definition.cables[0]
        wire = cable.wires[0]
        hrefs = list(sdn.get_hports(wire))
        self.assertTrue(len(hrefs) == 2 and all(isinstance(x.item, sdn.Port) for x in hrefs))
        hrefs = list(sdn.get_hports(cable))
        self.assertTrue(len(hrefs) == 2 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_from_outerpin(self):
        library = self.netlist.libraries[1]
        definition = library.definitions[0]
        instance = definition.children[0]
        outerpin = next(iter(instance.pins))
        hrefs = list(sdn.get_hports(outerpin))
        self.assertTrue(len(hrefs) == 1 and all(isinstance(x.item, sdn.Port) for x in hrefs))

    def test_bad_selection_type(self):
        self.assertRaises(TypeError, self.netlist.get_hports, selection="NOT_AN_OPTION")
        self.assertRaises(TypeError, self.netlist.get_hports, selection=None)

    def test_of_bad_instance(self):
        hrefs = list(sdn.get_hports(sdn.Instance()))
        self.assertTrue(len(hrefs) == 0)

    def test_through_hierarchy_again(self):
        netlist = sdn.Netlist()

        library = netlist.create_library()
        library.name = 'work'

        leaf_def = library.create_definition()
        leaf_def.name = 'leaf'
        leaf_port = leaf_def.create_port()
        leaf_port.name = 'I'
        leaf_port.create_pins(1)

        bottom_def = library.create_definition()
        bottom_def.name = 'bottom'
        bottom_port = bottom_def.create_port()
        bottom_port.name = 'I'
        bottom_port.create_pins(1)
        leaf_inst = bottom_def.create_child()
        leaf_inst.name = 'leaf'
        leaf_inst.reference = leaf_def
        bottom_cable = bottom_def.create_cable()
        bottom_cable.name = 'bottom_cable'
        bottom_wire = bottom_cable.create_wire()
        bottom_wire.connect_pin(bottom_port.pins[0])
        bottom_wire.connect_pin(leaf_inst.pins[leaf_port.pins[0]])

        bottom_floating_wire = bottom_cable.create_wire()

        middle_def = library.create_definition()
        middle_def.name = 'middle'
        middle_port = middle_def.create_port()
        middle_port.name = "I"
        middle_port.create_pin()
        bottom_inst = middle_def.create_child()
        bottom_inst.name = 'bottom'
        bottom_inst.reference = bottom_def
        middle_cable = middle_def.create_cable()
        middle_cable.name = "middle_cable"
        middle_wire = middle_cable.create_wire()
        middle_wire.connect_pin(middle_port.pins[0])
        middle_wire.connect_pin(bottom_inst.pins[bottom_port.pins[0]])

        middle_floating_wire = middle_cable.create_wire()

        top_def = library.create_definition()
        top_def.name = 'top'
        top_port = top_def.create_port()
        top_port.name = "I"
        top_port.create_pin()
        middle_inst = top_def.create_child()
        middle_inst.name = 'middle'
        middle_inst.reference = middle_def
        top_cable = top_def.create_cable()
        top_cable.name = "top_cable"
        top_wire = top_cable.create_wire()
        top_wire.connect_pin(top_port.pins[0])
        top_wire.connect_pin(middle_inst.pins[middle_port.pins[0]])

        top_floating_wire = top_cable.create_wire()

        top_instance = sdn.Instance()
        top_instance.name = 'top'
        top_instance.reference = top_def
        netlist.top_instance = top_instance

        #look at wire_name
        href = next(netlist.get_hports('I'))
        self.assertTrue('clock', href.name)
        hrefs = set(sdn.get_hports(netlist.top_instance, recursive=True))
        self.assertTrue(href in hrefs)

        hrefs = set(sdn.get_hports(middle_cable))
        href_top_pin = next(sdn.get_hports(top_port.pins[0]))
        href_middle_pin = next(sdn.get_hports(middle_port.pins[0]))
        href_bottom_pin = next(sdn.get_hports(bottom_port.pins[0]))
        self.assertTrue(href_top_pin not in hrefs and href_middle_pin in hrefs and href_bottom_pin in hrefs)

        hrefs = set(sdn.get_hports(top_def, recursive=True))
        self.assertTrue(href_top_pin in hrefs and href_middle_pin in hrefs and href_bottom_pin in hrefs)

        all_pins = set(netlist.get_hports())
        self.assertTrue(len(all_pins) == 1)

        all_pins = set(netlist.get_hports(recursive=True))
        self.assertTrue(len(all_pins) == 4)



from spydrnet.ir.outerpin import OuterPin
import spydrnet as sdn


class Comparer:

    def __init__(self, ir_orig, ir_composer):
        assert ir_orig and ir_composer, "Both orig_netlist and composer netlist must be specified"
        self.ir_orig = ir_orig
        self.ir_composer = ir_composer

    def run(self):
        self.compare()

    def compare(self):
        # try:
        assert self.get_identifier(self.ir_orig) == self.get_identifier(self.ir_composer), \
            "Environments do not have the same identifier"
        assert self.get_original_identifier(self.ir_orig) == self.get_original_identifier(self.ir_composer), \
            "Environments do not have the same original identifier"
        # if there is no top instance in either then this test passes.
        if (self.ir_composer.top_instance != None or self.ir_orig.top_instance != None):
            self.compare_instances(
                self.ir_orig.top_instance, self.ir_composer.top_instance)
        assert len(self.ir_orig.libraries) == len(self.ir_composer.libraries), \
            "Environments do not have the same number of libraries"
        # except Exception:
        #     import pdb; pdb.set_trace()
        for orig_library in self.ir_orig.libraries:
            if orig_library.name == None:
                # ports with no name are not compared
                print("WARNING: library with name == None exists and is not compared")
                continue
            else:
                patterns = orig_library.name
            composer_library = next(
                sdn.get_libraries(self.ir_composer, patterns))
            self.compare_libraries(orig_library, composer_library)

    def compare_libraries(self, library_orig, library_composer):
        # try:
        assert self.get_identifier(library_orig) == self.get_identifier(library_composer), \
            "Libraries do not have the same identifier"
        assert self.get_original_identifier(library_orig) == self.get_original_identifier(library_composer), \
            "Libraries do not have the same original identifier"
        assert len(library_orig.definitions) == len(library_composer.definitions), \
            "Libraries do not have the same amount of definitions " \
            + str(library_orig) + " " + str(len(library_orig.definitions)) + " " +\
            str(library_composer) + " " + \
            str(len(library_composer.definitions))
        # except Exception:
        #     import pdb; pdb.set_trace()
        for orig_definition in library_orig.definitions:
            if orig_definition.name == None:
                # ports with no name are not compared
                print(
                    "WARNING: definitions with name == None exist but are not compared")
                continue
            else:
                patterns = orig_definition.name
            composer_definition = next(
                sdn.get_definitions(library_composer, patterns))
            self.compare_definition(orig_definition, composer_definition)

    def compare_definition(self, definition_orig, definition_composer, check_identifier=True):
        if check_identifier:
            assert self.get_identifier(definition_orig) == self.get_identifier(definition_composer), \
                "Definitions do not have the same identifier"
            assert self.get_original_identifier(definition_orig) == self.get_original_identifier(definition_composer), \
                "Definitions do not have the same original identifier"

        assert len(definition_orig.ports) == len(definition_composer.ports), \
            "Definitions do not have the same number of ports"
        # for orig_port, composer_port in zip(definition_orig.ports, definition_composer.ports):
        #     self.compare_ports(orig_port, composer_port)
        # do a smarter compare
        for orig_port in definition_orig.ports:
            if orig_port.name == None:
                # ports with no name are not compared
                print("WARNING: ports with name == None exist and are not compared")
                continue
            else:
                patterns = orig_port.name
            composer_port = next(sdn.get_ports(definition_composer, patterns))
            self.compare_ports(orig_port, composer_port)

        assert len(definition_orig.cables) == len(definition_composer.cables), \
            "Definitions do not have the same number of cables "\
            + definition_orig.name + " " + str(len(definition_orig.cables))\
            + " " + definition_composer.name + " " + \
            str(len(definition_composer.cables))

        for orig_cable in definition_orig.cables:
            if orig_cable.name == None:
                # ports with no name are not compared
                print("WARNING: cables with name == None exist and are not compared")
                continue
            else:
                patterns = orig_cable.name
            composer_cable = next(sdn.get_cables(
                definition_composer, patterns))
            self.compare_cables(orig_cable, composer_cable)

        assert len(definition_orig.children) == len(definition_composer.children), \
            "Definitions do not have the same number of instances"
        # for orig_instance, composer_cable in zip(definition_orig.children, definition_composer.children):
        #     self.compare_instances(orig_instance, composer_cable)
        for orig_instance in definition_orig.children:
            if orig_instance.name == None:
                # ports with no name are not compared
                print("WARNING: children with name == None exist and are not compared")
                continue
            else:
                if orig_instance.name.startswith("SDN_Assignment_"):
                    # skip assignment statements for now
                    continue
                else:
                    patterns = orig_instance.name

            composer_instance = next(
                sdn.get_instances(definition_composer, patterns))

            self.compare_instances(orig_instance, composer_instance)

        # compare assignemnt statements
        pattern = "SDN_Assignment_*"
        composer_generator = sdn.get_instances(definition_composer, pattern)
        orig_generator = sdn.get_instances(definition_orig, pattern)
        # i just need to make sure that they both contain the same number of each width assignment
        composer_dict = dict()
        orig_dict = dict()

        ci = next(composer_generator, None)
        while ci is not None:
            num = ci.name.split("_")[3]
            if num in composer_dict:
                composer_dict[num] += 1
            else:
                composer_dict[num] = 1
            ci = next(composer_generator, None)

        oi = next(orig_generator, None)
        while oi is not None:
            num = oi.name.split("_")[3]
            if num in orig_dict:
                orig_dict[num] += 1
            else:
                orig_dict[num] = 1
            oi = next(orig_generator, None)

        for k in composer_dict.keys():
            assert k in orig_dict and orig_dict[k] == composer_dict[k], "there are a different number of " + str(
                k) + " width assignment"

    def compare_cables(self, cable_orig, cable_composer):
        assert self.get_identifier(cable_orig) == self.get_identifier(cable_composer), \
            "Cables do not have the same identifier"
        assert self.get_original_identifier(cable_orig) == self.get_original_identifier(cable_composer), \
            "Cables do not have the same original identifier"
        assert len(cable_orig.wires) == len(cable_composer.wires), \
            "Cables do not have the same number of wires " + cable_orig.name + " " + \
            str(len(cable_orig.wires)) + " " + cable_composer.name + \
            " " + str(len(cable_composer.wires))
        # zip is left here because the order of wires in cables matters.
        for orig_wire, composer_wire in zip(cable_orig.wires, cable_composer.wires):
            assert len(orig_wire.pins) == len(composer_wire.pins), \
                "wires connect to a different number of pins " + \
                orig_wire.cable.name + " " + composer_wire.cable.name
            for orig_pin, composer_pin in zip(orig_wire.pins, composer_wire.pins):
                assert type(orig_pin) == type(composer_pin), \
                    "pin types do not match up."
                if isinstance(orig_pin, OuterPin):
                    self.compare_outer_pins(orig_pin, composer_pin)
                else:
                    self.compare_inner_pins(orig_pin, composer_pin)

    def compare_outer_pins(self, pin_orig, pin_composer):
        assert pin_orig.instance.reference == pin_orig.inner_pin.port.definition and \
            pin_composer.instance.reference == pin_composer.inner_pin.port.definition, \
            "DRC failure, outer pin instance reference on associated pin definition not the same"
        # try:
        assert self.are_instances_equivalent(pin_orig.instance, pin_composer.instance), \
            "Net does not connect to a pin on the same instance"
        # except Exception:
        #     import pdb; pdb.set_trace()

        assert self.are_inner_pins_equivalent(pin_orig.inner_pin, pin_orig.inner_pin), \
            "Net does not connect the same pin"

    def compare_inner_pins(self, pin_orig, pin_composer):
        assert self.are_inner_pins_equivalent(pin_orig, pin_composer), \
            "Net does not connect the same pin"

    def are_instances_equivalent(self, orig_instance, composer_instance):
        orig_name = self.get_identifier(orig_instance)
        composer_name = self.get_identifier(composer_instance)
        # assignment instances are not very well kept when written out so just compare the width
        if orig_name.startswith("SDN_Assignment_") and composer_name.startswith("SDN_Assignment_"):
            orig_split = orig_name.split("_")
            composer_split = composer_name.split("_")
            assert orig_split[3] == composer_split[3], "the widths of the assignments are off " + \
                orig_split[3] + " " + composer_split[3]
        else:
            assert orig_name == composer_name, "Names are not the same " + \
                orig_name + " " + composer_name
        assert self.get_identifier(orig_instance.reference) == self.get_identifier(composer_instance.reference) and \
            self.get_identifier(orig_instance.reference.library) == \
            self.get_identifier(composer_instance.reference.library) and \
            self.get_identifier(orig_instance.parent) == self.get_identifier(composer_instance.parent) and \
            self.get_identifier(orig_instance.parent.library) == \
            self.get_identifier(composer_instance.parent.library), \
            'Instances are not equivalent ' + orig_instance.name + " " + composer_instance.name
        return True

    def are_inner_pins_equivalent(self, orig_pin, composer_pin):
        assert orig_pin.port.pins.index(orig_pin) == composer_pin.port.pins.index(composer_pin), \
            "Pin indices do not match"
        assert self.get_identifier(orig_pin.port) == self.get_identifier(composer_pin.port) and \
            self.get_identifier(orig_pin.port.definition) == self.get_identifier(composer_pin.port.definition) and \
            self.get_identifier(orig_pin.port.definition.library) == \
            self.get_identifier(composer_pin.port.definition.library), \
            "Pins are not from equivalent ports"
        return True

    def compare_ports(self, port_orig, port_composer):
        assert self.get_identifier(port_orig) == self.get_identifier(port_composer), \
            "Ports do not have the same identifier"
        assert self.get_original_identifier(port_orig) == self.get_original_identifier(port_composer), \
            "Ports do not have the same original identifier"

        assert port_orig.direction == port_composer.direction, \
            "Ports are not facing the same direction, " + \
            str(port_orig.direction) + " " + str(port_composer.direction)

        assert port_orig.is_array == port_composer.is_array, \
            "Ports Array mismatch"

        assert len(port_orig.pins) > 0 and len(port_composer.pins) > 0, \
            "DRC failure, ports should have at least one pin"

        assert len(port_orig.pins) == len(port_composer.pins), \
            "Ports do not have the same number of pins, " + port_orig.name + " " + \
            str(len(port_orig.pins)) + " " + port_composer.name + \
            " " + str(len(port_composer.pins))

        assert self.get_identifier(port_orig.definition) == self.get_identifier(port_composer.definition) and \
            self.get_identifier(port_orig.definition.library) == \
            self.get_identifier(port_composer.definition.library), \
            "Ports do not belong to an equivalent definition"

    def compare_instances(self, instances_orig, instances_composer):
        instances_orig_identifier = self.get_identifier(instances_orig)
        instances_composer_identifier = self.get_identifier(instances_composer)
        assert instances_orig_identifier == instances_composer_identifier, \
            "Instances do not have the same identifier, orig \"{}\" " \
            "composer \"{}\"".format(
                instances_orig_identifier, instances_composer_identifier)

        instances_orig_original_identifier = self.get_original_identifier(
            instances_orig)
        instances_composer_original_identifier = self.get_original_identifier(
            instances_composer)
        assert instances_orig_original_identifier == instances_composer_original_identifier, \
            "Instances do not have the same original identifier, orig " \
            "\"{}\" composer \"{}\"".format(
                instances_orig_original_identifier, instances_composer_original_identifier)

        assert (instances_orig.reference == None and instances_composer.reference == None) or \
            (self.get_identifier(instances_orig.reference) == self.get_identifier(instances_composer.reference) and
             self.get_identifier(instances_orig.reference.library) ==
             self.get_identifier(instances_composer.reference.library)), \
            "Instances do not have the same reference definition."

        if "EDIF.properties" in instances_orig:
            assert "EDIF.properties" in instances_composer, "Composer is missing properties"
            properties_orig = instances_orig["EDIF.properties"]
            properties_composer = instances_composer["EDIF.properties"]
            for x in range(len(properties_orig)):
                for key, value in properties_orig[x].items():
                    assert properties_orig[x][key] == properties_composer[x][key], \
                        "Instances do not have the same properties"

    @staticmethod
    def get_identifier(obj):
        if obj == None:
            return None
        return obj.name
        # if "EDIF.identifier" in obj:
        #     return obj["EDIF.identifier"]

    @staticmethod
    def get_original_identifier(obj):
        if obj == None:
            return None
        if "EDIF.original_identifier" in obj:
            return obj["EDIF.original_identifier"]



from spydrnet import FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Definition, Instance, Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_hinstances(obj, *args, **kwargs):
    """
    get_hinstances(obj, ...)

    Get hierarchical references to instances *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection of objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_instances(netlist, ...)` would
        return all of the instances *within* the provided definition that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    href_instances : generator
        The hierarchical references to instances associated with a particular object or collection of objects.
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_hinstances() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'recursive', 'filter', 'is_case', 'is_re'} for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    filter_func = kwargs.get('filter', lambda x: True)
    recursive = kwargs.get('recursive', False)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            object_collection = list(iter(obj))
        except TypeError:
            object_collection = [obj]
    else:
        object_collection = [obj]
    if all(isinstance(x, (HRef, FirstClassElement, InnerPin, OuterPin, Wire)) for x in object_collection) is False:
        raise TypeError("get_hinstances() supports all netlist related objects and hierarchical references or a "
                        "collection of theses as the object searched, unsupported object provided")

    if isinstance(patterns, str):
        patterns = (patterns,)
    assert isinstance(patterns, (FirstClassElement, InnerPin, OuterPin, Wire)) is False

    return _get_instances(object_collection, patterns, recursive, is_case, is_re, filter_func)


def _get_instances(object_collection, patterns, recursive, is_case, is_re, filter_func):
    for result in filter(filter_func, _get_instances_raw(object_collection, patterns, recursive, is_case, is_re)):
        yield result


def _get_instances_raw(object_collection, patterns, recursive, is_case, is_re):
    in_namemap = set()
    in_yield = set()
    namemap = dict()
    instance_search = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Netlist):
            top_instance = obj.top_instance
            if top_instance:
                href = HRef.from_parent_and_item(None, top_instance)
                _update_namemap(href, recursive, in_namemap, namemap)
        elif isinstance(obj, HRef):
            if obj.is_valid is False:
                continue
            item = obj.item
            if isinstance(item, Instance):
                _update_namemap(obj, recursive, in_namemap, namemap)
            elif isinstance(item, (Port, Cable)):
                hinstance = obj.parent
                if hinstance and hinstance not in in_yield:
                    in_yield.add(hinstance)
                    yield hinstance
            elif isinstance(item, (InnerPin, Wire)):
                hcable_or_hport = obj.parent
                if hcable_or_hport:
                    hinstance = hcable_or_hport.parent
                    if hinstance and hinstance not in in_yield:
                        in_yield.add(hinstance)
                        yield hinstance
        elif isinstance(obj, Library):
            object_collection += obj.definitions
        elif isinstance(obj, Definition):
            instance_search |= obj.references
        elif isinstance(obj, Instance):
            instance_search.add(obj)
        elif isinstance(obj, (Port, Cable)):
            definition = obj.definition
            if definition:
                object_collection.append(definition)
        elif isinstance(obj, InnerPin):
            port = obj.port
            if port:
                definition = port.definition
                if definition:
                    object_collection.append(definition)
        elif isinstance(obj, OuterPin):
            instance = obj.instance
            if instance:
                instance_search.add(instance)
        elif isinstance(obj, Wire):
            cable = obj.cable
            if cable:
                definition = cable.definition
                if definition:
                    object_collection.append(definition)

    if instance_search:
        for href in HRef.get_all_hrefs_of_instances(instance_search):
            if href not in in_yield:
                in_yield.add(href)
                yield href

    for href in in_yield:
        in_namemap.discard(href)

    if in_namemap:
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    for href in result:
                        if href in in_namemap:
                            in_namemap.remove(href)
                            yield href
            else:
                for name in namemap:
                    if _value_matches_pattern(name, pattern, is_case, is_re):
                        result = namemap[name]
                        for href in result:
                            if href in in_namemap:
                                in_namemap.remove(href)
                                yield href


def _update_namemap(href, recursive, found, namemap):
    currently_recursive = True
    search_stack = [(href, False)]
    name_stack = list()
    while search_stack:
        href, visited = search_stack.pop()
        if visited:
            name_stack.pop()
        else:
            search_stack.append((href, True))
            name_stack.append(href.item.name if href.item.name else '')
            if len(name_stack) > 1:
                hname = '/'.join(name_stack[1:])
                if hname not in namemap:
                    namemap[hname] = list()
                namemap[hname].append(href)
            if currently_recursive:
                currently_recursive = recursive
                item = href.item
                reference = item.reference
                if reference:
                    for child in reference.children:
                        href_child = HRef.from_parent_and_item(href, child)
                        if href_child not in found:
                            found.add(href_child)
                            search_stack.append((href_child, False))



# Copyright 2020 Dallin Skouson, Andrew Keller, Michael Wirthlin

from collections import deque
# from spydrnet.ir import *

"""Code to make definitions unique throughout a netlist.
expected parameters,
uniqify -- Makes all definitions unique below the top instance. definitions that are not referenced below the top instance will not be unique.
"""


MOD_NAME_UID = 0


def _get_unique_name_modifier():
    global MOD_NAME_UID
    str_out = "_sdn_unique_" + str(MOD_NAME_UID)
    MOD_NAME_UID += 1
    return str_out


def _make_instance_unique(instance):
    """clone the definition and point the reference to the new definition"""
    reference = instance.reference
    lib = instance.reference.library
    index = lib.definitions.index(reference)
    new_def = instance.reference.clone()
    if instance.reference.name is not None:
        name = instance.reference.name
        unique_suffix = _get_unique_name_modifier()
        new_def.name = name + unique_suffix
        if 'EDIF.identifier' in new_def:
            new_def['EDIF.identifier'] = new_def['EDIF.identifier'] + \
                unique_suffix
    lib.add_definition(new_def, index + 1)
    instance.reference = new_def


def _is_unique(instance):
    """Return if the instance is the only one of its kind in the definition"""
    return len(instance.reference.references) == 1 or instance.reference.is_leaf()


def uniquify(netlist):
    """Make the instances in the netlist unique
    uniqification is done in place. Each instance will correspond to exactly one definition and each definition will correspond to exactly one instance with the exception of leaf cells.
    Leaf cells are can be instanced unlimited numbers of times. Any netlist elements that are not instantiated by the top instance will not be modified and may retain duplicate instances
    Currently there is no guarantee that the original definition names will be maintained, but it is guaranteed that they will be unique within the scope of all hardware that is below the top instance.

    Renaming is predictable. the string: _sdn_unique_# will be added to the end of the definition names.

    :param netlist: the netlist that will be uniquified

    :return: void

    """

    # import pdb; pdb.set_trace()
    # starting with top.
    # top must be unique below top. otherwise we have infinite harware recursion which is does not make much sense.
    instance_queue = deque()

    top_instance = netlist.top_instance

    top_definition = top_instance.reference

    # put all of tops children on a stack
    for chld in top_definition.children:
        instance_queue.append(chld)

    # for each of the children on the stack
    while len(instance_queue) > 0:
        inst = instance_queue.popleft()
        # if they are not unique
        if not _is_unique(inst):
            # uniquify
            _make_instance_unique(inst)
        # put their children on the stack
        for chld in inst.reference.children:
            instance_queue.append(chld)



import os


def compose(netlist, filename, definition_list=[], write_blackbox=True):
    """To compose a file into a netlist format"""
    extension = os.path.splitext(filename)[1]
    extension_lower = extension.lower()
    if extension_lower in {".edf", ".edif"}:
        from spydrnet.composers.edif.composer import ComposeEdif
        composer = ComposeEdif()
        if netlist.name is None:
            raise Exception("netlist.name undefined")
        composer.run(netlist, filename)
    elif extension_lower in [".v", ".vh"]:
        from spydrnet.composers.verilog.composer import Composer
        composer = Composer(definition_list, write_blackbox)
        composer.run(netlist, file_out=filename)
    else:
        raise RuntimeError("Extension {} not recognized.".format(extension))



import spydrnet.ir as ir
import weakref

from spydrnet.shortcuts.getter import GetterShortcuts

flyweight = weakref.WeakKeyDictionary()


class HRef(GetterShortcuts):
    """A hierarchical reference to a specific element in a netlist.

    Definitions can be instanced more than once (i.e., multiple instances can reference the same definition).
    When a definition is instanced more than once, it causes the contents of the definition to be shared. Therefore, any
    changes to a multi-instanced definition will be reflected in all instances of that definition. Similarly, any
    references to the contents of a multi-instanced definition refer to the contents of all of the instances and not to
    the contents of a specific instance. This sharing creates challenges for analyzing and transforming the netlist.

    Hierarchical references refer to a netlist element by hierarchical sequence. A hierarchical sequence begins with the
    the top-instance of netlist (see Netlist.top_instance). The sequence continues with children instances (parent to
    child) until the instance of interest is reached. The instance of interest is the final instance in the sequence.
    When the referenced element is an instance, the sequence terminates. When the referenced element is a port, pin,
    cable, or wire, the sequence continues with those elements until the desired element is specified (e.g., port;
    port, pin; cable; or cable, wire. In this way, hierarhical elements are uniquely referenced even though the contents
    of a definition may be shared.

    **Hierarchical Sequence Examples:**

    Here are some examples of hierarchical sequences:

    * Top Instance:
        * ``[top_instance]``
    * Top Instance Port
        * ``[top_instance, port]``
    * Top Instance Pin
        * ``[top_instance, port, pin]``
    * Shared Sub-Instance Cable)
        * ``[top_instance, sub_instance_A, sub_instance_C, cable]``
        * ``[top_instance, sub_instance_B, sub_instance_C, cable]``
        * ``sub_instance_A`` and ``sub_instance_B`` are instances (or children) with the definition referenced by
          ``top_instance``.
        * ``sub_instance_A`` and ``sub_instance_B`` reference the same definition, which contains ``sub_instance_C``.
        * Even though ``cable`` is the same element in both sequences, each sequence uniquely references the cable
          inside ``sub_instance_A`` and ``sub_instance_B`` respectively.

    **Netlist Analysis and Transformation:**

    Hierarchical references provide unique handles on hierarchical elements. A unique handle allows for such elements to
    be considered individually even though two hierarchical may point to some of the same elements. This makes it
    possible, for example, to consider pin connectivity across hierarchy even though the actual pins may be the same.

    In some netlist tranformations, it may be desirable to modify the contents of a specific instance without modifying
    the contents of another instance that refers to the same definition. Hierarchical references make it possible to
    refer to an instance that should be changed. Once the definition is made unique (see spydrnet.uniquify), then any
    alterations will only affect the originally specified instance. Hierarchical instances also allow for uniqueness
    checking (see HRef.is_unique).

    **Hierarchical Reference Representation:**

    HRefs represent hierarchy as nodes in a hierarchical tree. The root node is an HRef to the top_instance of a netlist
    with no parent node. Each HRef contains a pointer to its parent HRef (``None`` in the case of the root HRef), a
    pointer to the element in the netlist that it references, and a hashcode generated from each referenced object.

    Storing the hashcode with the object saves on re-computation and allows for quick operations in containers that
    require Hashable objects. If the hashcode were not stored with the object, it would have to be recalculated for
    each hash-dependent operation, which could consume a large amount of computational resources depending on the
    hierarchical depth of the node. Parent and item pointers are immutable. The hashcode of a referenced item is also
    immutable. Therefore, the hashcode of a HRef should not change during its existence (even if a netlist
    transformation renders it invalid).

    **Use of a Flyweight Pattern:**

    Due the the nature of hierarchical references, parent nodes can be referenced more than once. Rather than having
    multiple hierarchical nodes in memory that point to the same hierarchical parent, a flyweight can be used to save
    on memory. A flyweight pattern is used here to share hierarchical parent nodes. See `Flyweight pattern
    <https://en.wikipedia.org/wiki/Flyweight_pattern>`_.

    **Lack of Parent to Child Pointers:**

    A parent to child pointer requires a lookup diction from each child item to each child hierarchical node. This
    approach could be taken, but it recreates much of the same information that is available in the original netlist. It
    was therefore decided to leverage the flyweight pattern rather than explicitly manage all of the necessary
    child-item to child-node relationships.

    .. attribute:: item: the item of the object
    .. attribute:: parent: the parent of the object
    """
    @staticmethod
    def get_all_hrefs_of_item(item):
        """Get all the href of the itsm

        parameters
        ----------

        item - The item to get the href from.

        """
        if isinstance(item, ir.Instance):
            for href in HRef.get_all_hrefs_of_instances(item):
                yield href
        elif isinstance(item, ir.Definition):
            for href in HRef.get_all_hrefs_of_instances(item.references):
                yield href
        elif isinstance(item, ir.Port):
            definition = item.definition
            if definition:
                for href in HRef.get_all_hrefs_of_instances(definition.references):
                    yield HRef.from_parent_and_item(href, item)
        elif isinstance(item, ir.InnerPin):
            port = item.port
            if port:
                definition = port.definition
                if definition:
                    for href in HRef.get_all_hrefs_of_instances(definition.references):
                        href_port = HRef.from_parent_and_item(href, port)
                        yield HRef.from_parent_and_item(href_port, item)
        elif isinstance(item, ir.OuterPin):
            instance = item.instance
            inner_pin = item.inner_pin
            if instance and inner_pin:
                port = inner_pin.port
                if port:
                    for href in HRef.get_all_hrefs_of_instances(instance):
                        href_port = HRef.from_parent_and_item(href, port)
                        yield HRef.from_parent_and_item(href_port, inner_pin)
        elif isinstance(item, ir.Cable):
            definition = item.definition
            if definition:
                for href in HRef.get_all_hrefs_of_instances(definition.references):
                    yield HRef.from_parent_and_item(href, item)
        elif isinstance(item, ir.Wire):
            cable = item.cable
            if cable:
                definition = cable.definition
                if definition:
                    for href in HRef.get_all_hrefs_of_instances(definition.references):
                        href_cable = HRef.from_parent_and_item(href, cable)
                        yield HRef.from_parent_and_item(href_cable, item)

    @staticmethod
    def get_all_hrefs_of_instances(instances, netlist=None):
        """
        Assuming all instances are vaild (meaning their reference belongs in a proper library inside a netlist).
        :param instances:
        :param netlist:
        :return:
        """
        if isinstance(instances, ir.Instance):
            instances = {instances}
        else:
            instances = set(x for x in instances if isinstance(x, ir.Instance))

        if netlist is None:
            instance = next(iter(instances), None)
            if instance:
                reference = instance.reference
                if reference:
                    library = reference.library
                    if library:
                        netlist = library.netlist
        if isinstance(netlist, ir.Netlist) is False:
            return

        top_instance = netlist.top_instance

        bound = set()
        search_stack = list(instances)
        while search_stack:
            instance = search_stack.pop()
            parent_def = instance.parent
            if parent_def:
                for parent_inst in parent_def.references:
                    if parent_inst not in bound:
                        bound.add(parent_inst)
                        search_stack.append(parent_inst)

        href = HRef.from_parent_and_item(None, top_instance)
        search_stack = [href]
        while search_stack:
            href = search_stack.pop()
            item = href.item
            if item in instances:
                yield href
            reference = item.reference
            if reference:
                for child in reference.children:
                    if child in bound:
                        href_child = HRef.from_parent_and_item(href, child)
                        search_stack.append(href_child)
                    elif child in instances:
                        href_child = HRef.from_parent_and_item(href, child)
                        yield href_child

    @staticmethod
    def from_sequence(sequence):
        """Return the href of the sequence

        parameters
        ----------

        sequence - The sequence to get the href from

        """
        parent = None
        for item in sequence:
            href = HRef.from_parent_and_item(parent, item)
            parent = href
        return href

    @staticmethod
    def from_parent_and_item(parent, item):
        """Return the href with given parent and item

        parameters
        ----------

        parent - the parent obejct of this href
        item - the item that the href is reference to

        """
        href = HRef(item, parent)
        if href in flyweight:
            return flyweight[href]()
        else:
            flyweight[href] = weakref.ref(href)
            return href

    __slots__ = ['_hashcode', 'parent', 'item', '__weakref__']

    def __init__(self, item, parent=None):
        """Initialize the href

        parameters
        ----------

        item - the item that the href is reference to
        parent - the parent obejct of this href

        """
        self._hashcode = hash(hash(parent)*31 + hash(item))
        self.parent = parent
        self.item = item

    def __hash__(self):
        return self._hashcode

    def __eq__(self, other):
        if self is other:
            return True
        if isinstance(other, HRef) is False:
            return False
        this = self
        that = other
        while this is not None and that is not None:
            if this.item == that.item:
                this = this.parent
                that = that.parent
            else:
                return False
        if this is None and that is None:
            return True
        return False

    def __repr__(self):
        return "<{} {} '{}' at 0x{:016X}>".format(self.__class__.__name__, self.item.__class__.__name__, self.name,
                                                  id(self))

    def __str__(self):
        name = self.name
        return name

    @property
    def is_unique(self):
        """
        A hierarchical reference must be valid to be unique. If it is not valid, it may not be unique.
        : return:
        """
        if self.is_valid is False:
            return False
        href = self
        search_stack = list()
        instances = set()
        while href:
            item = href.item
            if isinstance(item, ir.Instance):
                instances.add(item)
                parent = item.parent
                if parent and len(parent.references) > 1 and href.parent:
                    search_stack += (x for x in parent.references if x !=
                                     href.parent.item)
            href = href.parent
        while search_stack:
            instance = search_stack.pop()
            reference = instance.parent
            if reference:
                for parent in reference.references:
                    if parent in instances:
                        return False
                    else:
                        search_stack.append(parent)
        return True

    @property
    def is_valid(self):
        """Checks if the href is valid

        """
        href = self
        while href:
            hparent = href.parent  # href
            item = href.item
            if isinstance(item, ir.Instance):
                if not hparent:
                    reference = item.reference
                    if not reference:
                        return False
                    library = reference.library
                    if not library:
                        return False
                    netlist = library.netlist
                    if not netlist:
                        return False
                    top_instance = netlist.top_instance
                    if not top_instance:
                        return False
                    if top_instance == item:
                        return True
                    return False
                else:
                    parent = item.parent  # definition
                    if not parent:
                        return False
                    hparent_item = hparent.item  # instance
                    if hparent_item not in parent.references:
                        return False
                    href = hparent
            elif isinstance(item, (ir.Cable, ir.Port)):
                if not hparent:
                    return False
                definition = item.definition
                if not definition:
                    return False
                if hparent.item not in definition.references:
                    return False
                href = hparent
            elif isinstance(item, ir.Wire):
                if not hparent:
                    return False
                cable = item.cable
                if not cable:
                    return False
                if hparent.item != cable:
                    return False
                href = hparent
            elif isinstance(item, ir.InnerPin):
                if not hparent:
                    return False
                port = item.port
                if not port:
                    return False
                if hparent.item != port:
                    return False
                href = hparent
            else:
                href = None
        return False

    @property
    def name(self):
        """Stores the name of the href

        """
        hseperator = '/'
        names = list()
        index = None
        item = self.item
        if isinstance(item, ir.Wire):
            cable = item.cable
            if cable.is_array:
                index = cable.lower_index + cable.wires.index(item)
            href = self.parent
        elif isinstance(item, ir.InnerPin):
            port = item.port
            if port.is_array:
                index = port.lower_index + port.pins.index(item)
            href = self.parent
        else:
            href = self
        while href is not None:
            item = href.item
            item_name = item.get(".NAME", "")
            names.append(item_name)
            href = href.parent
        bus_index = "" if index is None else "[{}]".format(index)
        hname = "{}{}".format(hseperator.join(reversed(names[:-1])), bus_index)
        return hname



from spydrnet.ir.pin import Pin
from copy import deepcopy, copy, error


class OuterPin(Pin):
    """
    Pins that correspond to instances. These pins can be thought of as on the outside of an instance. There can be many
    outer pins for each inner pin
    """
    __slots__ = ['_instance', '_inner_pin']

    @staticmethod
    def from_instance_and_inner_pin(instance, inner_pin):
        """Create an outer pin associated with a given inner_pin and instance object.

        parameters
        ----------

        instance - (Instance) the instance to associate with this pin

        inner_pin - (InnerPin) the inner pin with which to associate this outer pin"""
        return OuterPin(instance, inner_pin)

    def __init__(self, instance=None, inner_pin=None):
        """create an OuterPin.

        parameters
        ----------

        instance - (Instance) the instance with which to associate this outper pin.

        inner_pin - (InnerPin) a definition's inner pin to be associated with this instance outer pin."""
        super().__init__()
        self._instance = instance
        self._inner_pin = inner_pin

    @property
    def instance(self):
        """Return the instance with which this pin is associated"""
        return self._instance

    @property
    def inner_pin(self):
        """get the inner pin associated with this outer pin"""
        return self._inner_pin

    def __eq__(self, other):
        if isinstance(other, OuterPin):
            return self._instance == other._instance and self._inner_pin == other._inner_pin
        return False

    def __hash__(self):
        return hash((self._instance, self._inner_pin))

    def _clone_rip_and_replace(self, memo):
        """remove from its current environment and place it into the new cloned environment with references held in the memo dictionary"""
        if self._wire != None:
            assert self._wire in memo, "can't call this function when the wire has not been cloned yet"
            self._wire = memo[self._wire]

    def _clone_rip(self):
        """Remove from its current environmnet. 

        This will remove all pin pointers and create a floating stand alone instance.
        """
        self._inner_pin = None
        self._wire = None

    def _clone(self, memo):
        """Not api safe clone function

        clone leaving all references in tact.
        the element can then either be ripped or ripped and replaced"""
        assert self not in memo, "the object should not have been copied twice in this pass"
        c = OuterPin()
        memo[self] = c
        c._instance = None
        c._inner_pin = self._inner_pin
        c._wire = self._wire
        return c

    def clone(self):
        """Clone the pin in an api safe way.

        The following conditions will be met with the returned outer pin:

         * the pin will not be connected to any wires
         * the pin will be orphaned from any instance
         * the pin will not be connected to any inner pins
        """
        c = self._clone(dict())
        c._clone_rip()
        return c



from spydrnet.ir import FirstClassElement, InnerPin, OuterPin, Wire, Netlist, Library, Element, Definition, Instance, \
    Port, Cable
from spydrnet.util.hierarchical_reference import HRef
from spydrnet.util.selection import Selection
from spydrnet.global_state.global_service import lookup
from spydrnet.util.patterns import _is_pattern_absolute, _value_matches_pattern


def get_libraries(obj, *args, **kwargs):
    """
    get_libraries(obj, ...)

    Get libraries *within* an object.

    Parameters
    ----------
    obj : object, Iterable - required
        The object or objects associated with this query. Queries return a collection objects associated with the
        provided object or objects that match the query criteria. For example, `sdn.get_libraries(netlist, ...)` would
        return all of the libraries associated with the provided netlist that match the additional criteria.
    patterns : str, Iterable - optional, positional or named, default: wildcard
        The search patterns. Patterns can be a single string or an Iterable collection of strings. Patterns can be
        absolute or they can contain wildcards or regular expressions. If `patterns` is not provided, then it defaults
        to a wildcard. Patterns are queried against the object property value stored under a specified key. Fast lookups
        are only attempted on absolute patterns that are not regular expressions and contain no wildcards.
    key : str, optional, default: ".NAME"
        This is the key that controls which value is being searched.
    is_case : bool - optional, named, default: True
        Specify if patterns should be treated as case sensitive. Only applies to patterns. Does not alter fast lookup
        behavior (if namespace policy uses case insensitive indexing, this parameter will not prevent a fast lookup
        from returning a matching object even if the case is not an exact match).
    is_re: bool - optional, named, default: False
        Specify if patterns are regular expressions. If `False`, a pattern can still contain `*` and `?` wildcards. A
        `*` matches zero or more characters. A `?` matches upto a single character.
    recursive : bool - optional, default: False
        Specify if search should be recursive or not meaning that sub hierarchical instances within an instance are
        included or not.
    selection : Selection.{INSIDE, OUTSIDE}, default: INSIDE
        This parameter determines the instances that are returned based on the definition that is being searched. This
        parameter only applies to objects that are definitions. If the selection is "INSIDE" (default), then the
        function will return all of the instances that are inside the definition (i.e., the definition's children) that
        match the remainder of the search criteria. If the selection is "OUTSIDE", then the function will return all of
        the instances of the provided definition that match the remainder of the search criteria.
    filter : function
        This is a single input function that can be used to filter out unwanted virtual instances. If not specifed, all
        matching virtual instances are returned. Otherwise, virtual instances that cause the filter function to evaluate
        to true are the only items returned.
    
    Returns
    -------
    libraries : generator
        The libraries associated with a particular object
    
    """
    # Check argument list
    if len(args) == 1 and 'patterns' in kwargs:
        raise TypeError("get_libraries() got multiple values for argument 'patterns'")
    if len(args) > 1 or any(x not in {'patterns', 'key', 'filter', 'is_case', 'is_re', 'selection', 'recursive'}
                            for x in kwargs):
        raise TypeError("Unknown usage. Please see help for more information.")

    # Default values
    selection = kwargs.get('selection', Selection.INSIDE)
    if isinstance(selection, str):
        if selection in Selection.__members__:
            selection = Selection[selection]
    if selection not in {Selection.INSIDE, Selection.OUTSIDE}:
        raise TypeError("selection must be '{}'".format("', '".join([Selection.INSIDE.name, Selection.OUTSIDE.name])))

    filter_func = kwargs.get('filter', lambda x: True)
    is_case = kwargs.get('is_case', True)
    is_re = kwargs.get('is_re', False)
    patterns = args[0] if len(args) == 1 else kwargs.get('patterns', ".*" if is_re else "*")
    key = kwargs.get('key', ".NAME")
    recursive = kwargs.get('recursive', False)

    if isinstance(obj, (FirstClassElement, InnerPin, OuterPin, Wire)) is False:
        try:
            netlist_collection = list(iter(obj))
        except TypeError:
            netlist_collection = [obj]
    else:
        netlist_collection = [obj]
    if all(isinstance(x, (Element, HRef)) for x in netlist_collection) is False:
        raise TypeError("get_libraries() supports all netlist elements and hierarchical references or a collection of "
                        "theses as the object searched, unsupported object provided")

    if isinstance(patterns, str):
        patterns = (patterns,)

    return _get_libraries(netlist_collection, patterns, key, is_case, is_re, selection, recursive, filter_func)


def _get_libraries(netlist_collection, patterns, key, is_case, is_re, selection, recursive, filter_func):
    for result in filter(filter_func, _get_libraries_raw(netlist_collection, patterns, key, is_case, is_re, selection,
                                                         recursive)):
        yield result


def _get_libraries_raw(object_collection, patterns, key, is_case, is_re, selection, recursive):
    found = set()
    other_libraries = set()
    while object_collection:
        obj = object_collection.pop()
        if isinstance(obj, Netlist):
            for pattern in patterns:
                pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
                if pattern_is_absolute:
                    result = lookup(obj, Library, key, pattern)
                    if result is not None and result not in found:
                        found.add(result)
                        yield result
                else:
                    for library in obj.libraries:
                        value = library[key] if key in library else ''
                        if library not in found and _value_matches_pattern(value, pattern, is_case, is_re):
                            found.add(library)
                            yield library
        elif isinstance(obj, Library):
            # Libraries used by this library if inside
            if selection == Selection.INSIDE:
                for definition in obj.definitions:
                    for child in definition.children:
                        reference = child.reference
                        if reference:
                            library = reference.library
                            if library and library not in other_libraries:
                                other_libraries.add(library)
                                if recursive:
                                    object_collection.append(library)
            # Libraries that use this library if outside
            else:
                for definition in obj.definitions:
                    for instance in definition.references:
                        parent = instance.parent
                        if parent:
                            library = parent.library
                            if library and library not in other_libraries:
                                other_libraries.add(library)
                                if recursive:
                                    object_collection.append(library)
        elif isinstance(obj, Definition):
            if selection == Selection.INSIDE:
                library = obj.library
                if library and library not in other_libraries:
                    other_libraries.add(library)
                if recursive:
                    object_collection += obj.children
            else:
                for instance in obj.references:
                    parent = instance.parent
                    if parent:
                        library = parent.library
                        if library and library not in other_libraries:
                            other_libraries.add(library)
                        if recursive:
                            object_collection.append(parent)
        elif isinstance(obj, Instance):
            if selection == Selection.INSIDE:
                reference = obj.reference
                if reference:
                    library = reference.library
                    if library and library not in other_libraries:
                        other_libraries.add(library)
                    if recursive:
                        object_collection += reference.children
            else:
                parent = obj.parent
                if parent:
                    library = parent.library
                    if library and library not in other_libraries:
                        other_libraries.add(library)
                    if recursive:
                        object_collection += parent
        elif isinstance(obj, (Port, Cable)):
            definition = obj.definition
            if definition:
                object_collection.append(definition)
        elif isinstance(obj, InnerPin):
            port = obj.port
            if port:
                object_collection.append(port)
        elif isinstance(obj, OuterPin):
            instance = obj.instance
            if instance:
                object_collection.append(instance)
        elif isinstance(obj, Wire):
            cable = obj.cable
            if cable:
                object_collection.append(cable)
        elif isinstance(obj, HRef):
            if obj.is_valid:
                object_collection.append(obj.item)

    if other_libraries:
        namemap = dict()
        for other_instance in other_libraries:
            if other_instance in found:
                continue
            found.add(other_instance)
            name = other_instance[key] if key in other_instance else ''
            if name not in namemap:
                namemap[name] = list()
            namemap[name].append(other_instance)
        for pattern in patterns:
            pattern_is_absolute = _is_pattern_absolute(pattern, is_case, is_re)
            if pattern_is_absolute:
                if pattern in namemap:
                    result = namemap[pattern]
                    for instance in result:
                        yield instance
            else:
                discard = set()
                for instance in found:
                    value = instance[key] if key in instance else ''
                    if _value_matches_pattern(value, pattern, is_case, is_re):
                        discard.add(instance)
                        yield instance
                found -= discard



import unittest
import spydrnet as sdn

from spydrnet.parsers.edif.parser import EdifParser
from spydrnet import base_dir
import os
import tempfile
import glob
import shutil


class TestEdifTokenizer(unittest.TestCase):
    def test_multi_bit_add_out_of_order(self):
        definition = sdn.Definition()
        cable0 = sdn.Cable()
        cable0.name = "net[0]"
        cable0["EDIF.identifier"] = "net_0_"
        cable1 = sdn.Cable()
        cable1.name = "net[1]"
        cable1["EDIF.identifier"] = "net_1_"
        cable2 = sdn.Cable()
        cable2.name = "net[2]"
        cable2["EDIF.identifier"] = "net_2_"

        cable0.create_wire()
        cable1.create_wire()
        cable2.create_wire()
        
        p0 = sdn.InnerPin()
        p1 = sdn.InnerPin()
        p2 = sdn.InnerPin()

        cable1.wires[0].connect_pin(p0)
        cable1.wires[0].connect_pin(p1)
        cable1.wires[0].connect_pin(p2)

        ep = EdifParser()
        ep.multibit_add_cable(definition, cable0)
        ep.multibit_add_cable(definition, cable2)
        ep.multibit_add_cable(definition, cable1)

        assert len(definition.cables) == 1
        assert len (definition.cables[0].wires) == 3
        assert len(definition.cables[0].wires[0].pins) == 0
        assert len(definition.cables[0].wires[1].pins) == 3
        assert len(definition.cables[0].wires[2].pins) == 0
        assert p0 in definition.cables[0].wires[1].pins
        assert p1 in definition.cables[0].wires[1].pins
        assert p2 in definition.cables[0].wires[1].pins
            

    @classmethod
    def setUpClass(cls) -> None:
        cls.dir_of_edif_netlists = os.path.join(sdn.base_dir, "support_files", "EDIF_netlists")
        cls.edif_files = sorted(glob.glob(os.path.join(cls.dir_of_edif_netlists, "*.edf.zip")), key=os.path.getsize)

    @unittest.skip("Test takes a long time right now.")
    def test_large_edif(self):
        for ii, filename in enumerate(self.edif_files):
            if os.path.getsize(filename) <= 1024 * 10:
                continue
            self.ensure_cable_consistency(filename, ii, "edf")

    def test_small_edif_cables(self):
        for ii, filename in enumerate(self.edif_files):
            if os.path.getsize(filename) > 1024 * 10:
                continue
            self.ensure_cable_consistency(filename, ii, "edf")

    def ensure_cable_consistency(self,filename, ii, target_format_extension = None):
        with self.subTest(i=ii):
            if os.path.exists("temp"):
                shutil.rmtree("temp")
            print(filename)
            with tempfile.TemporaryDirectory() as tempdirname:
                netlist = sdn.parse(filename)
                for l in netlist.libraries:
                    for d in l.definitions:
                        for c in d.cables:
                            assert c.definition is not None


if __name__ == '__main__':
    unittest.main()
